/*
 * File:   GPS406a.c
 * Author: Daniel Hankewycz
 * Description: These functions are written for the Penn State Buss Tracking
 *              project. These functions provides methods to parse and interpret
 *              an incomming string from a EM-406A GPS module.
 *
 * Created on May 8, 2013, 1:06 AM
 */
#include <p18f26k22.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "GPS406a.h"
#include "myPic.h"

static int16 errorQueue[15];
static int8 errorQueueSize = 0;
/*
 * ISR ERRORS: 0x10-0x1f
 *      0x10: Tried to flip buffer pointers, but main is not done reading from buffer
 *      0x11: buffIndex has exceeded 100 characters
 * PARSING ERRORS: 0x20-0x2f
 *      0x20: modeName does not satisfy any of the given data stream formats
 *      0x21: bad checksum
 *      0x22: invalid coordinate (insufficient satelite lock count)
 */

//----------------------------------------------
// Function Implementations
//----------------------------------------------

//-----------------------------------------------------------------------------
// GETMODE
// Given: a properly tokenized string containing GPS data
// Return: the mode of data
//_______TABLE OF MODES____________
//  1: GGA
//  2: GSA
//  3: GSV
//  4: RMC
//  5: VTG
//  6: GLL
//-----------------------------------------------------------------------------

int16 getMode(char line[]) {

    int16 result = 0;
    char modeName[6];
    int8 size; //length of line
    int8 n = 0; //index used when writing modeName
    int8 i = 0; //index used to iterate through line

    size = strLength(line);
	if (size >= MAX_STRING_LENGTH-1) return(-1);

    // move index upto the '$' character
    while ((line[i++] != '$') && (i < size));

    while ((i < size) && (line[i] != ','))
        modeName[n++] = line[i++];
    modeName[5] = '\0';

    if (strcmppgm2ram(modeName, (far rom char *) "GPGGA") == 0) {
        result = GPGGA;
    } else if (strcmppgm2ram(modeName, (far rom char *) "GPGSA") == 0) {
        result = GPGSA;
    } else if (strcmppgm2ram(modeName, (far rom char *) "GPGSV") == 0) {
        result = GPGSV;
    } else if (strcmppgm2ram(modeName, (far rom char *) "GPRMC") == 0) {
        result = GPRMC;
    } else if (strcmppgm2ram(modeName, (far rom char *) "GPVTG") == 0) {
        result = GPVTG;
    } else if (strcmppgm2ram(modeName, (far rom char *) "GPGLL") == 0) {
        result = GPGLL;
    } else {
        // modeName does not satisfy any data stream format
        pushError(0x20);
    }// if

    return result;
}// getMode()


//-----------------------------------------------------------------------------
// PARSEDATARMC
// Given: pointer to a properly formated RMC char array, time array,
//			longitude array, longitude array
// Result: Returns the time, latitude, longitude, velocity, and heading data
//          from the RMC string into their respective arrays
//
//	Name			Example 			Units/Description
//	Message ID 		$GPRMC 				RMC protocol header
//	UTC Time 		161229.487 			hhmmss.sss
//	Status 			A 					A=data valid or V=data not valid
//	Latitude 		3723.2475 			ddmm.mmmm
//	N/S Indicator 	N 					N=north or S=south
//	Longitude 		12158.3416 			dddmm.mmmm
//	E/W Indicator 	W 					E=east or W=west
//	Speed Over Gnd	0.13				knots
//	Crse Over Gnd	309.62				degrees True
//	Date			120598				ddmmyy
//	Magnetic Varr	2 					degrees E=east or W=west
//	Checksum 		*10
//	<CR><LF> 							End of message termination
//	SiRF Technology Inc. does not support magnetic declination. All course over ground data
//-----------------------------------------------------------------------------

void parseDataRMC(char in[]) {

    int8 i = 0;
    int8 t = 0;
    int8 size;
	char scratch[10];

    time[0] = 0x00;
    latitude[0] = 0x00;
    longitude[0] = 0x00;

    size = strLength(in);
	if (size >= MAX_STRING_LENGTH-1) return;

    if (checkSum(in)) {

        //Skip the message ID token
        while ((in[i] != ',') && (i < size)) i++;

        //store the UTC time token 
        i++;
        while ((in[i] != ',') && (t < 11) && (i < size)) 
			if ((in[i] >= ' ') && (in[i] < 'z'))	time[t++] = in[i++];
			else i += 1;
        time[t] = '\0';

        //Store the data valid token
        i++;
        while ((in[i] != ',') && (i < size))
			if (in[i++] != 'A')	gpsLock = FALSE;
			else 				gpsLock = TRUE;
            

        //GET THE LATITUDE TOKEN
        i++;
        t = 1; // leave space for N/S indicator
        while ((in[i] != ',') && (t < 11) && (i < size))
			if ((in[i] >= ' ') && (in[i] < 'z')) latitude[t++] = in[i++];
			else i += 1;
        latitude[t] = '\0';


        //GET THE N/S INDICATOR TOKEN
        i++;
        while ((in[i] != ',') && (i < size)) {
            if (in[i] == 'S')	latitude[0] = '-';
            else                latitude[0] = ' ';
            i++;
        }// while

        //GET THE LONGITUDE TOKEN
        i++;
        t = 1; // leave space for N/S indicator
        while ((in[i] != ',') && (t < 11) && (i < size))	
			if ((in[i] >= ' ') && (in[i] < 'z')) longitude[t++] = in[i++];
			else i += 1;
        longitude[t] = '\0';

        //GET THE E/W INDICATOR TOKEN
        i++;
        while ((in[i] != ',') && (i < size)) {
            if (in[i] == 'W')	longitude[0] = '-';
            else                longitude[0] = ' ';
            i++;
        }// while

        //GET THE VELOCITY TOKEN
        i++;
        t = 0; // leave space for E/W indicator
        while ((in[i] != ',') && (in[i] != '.') && (t < 5) && (i < size))
		    if ((in[i] >= ' ') && (in[i] < 'z')) velocity[t++] = in[i++];
			else i+=1;
		velocity[t] = '\0';
		while ((in[i] != ',') && (i < size)) i += 1;

        //GET THE HEADING TOKEN
        i++;
        t = 0;
        while ((in[i] != ',') && (in[i] != '.') && (t < 7) && (i < size))		
		    if ((in[i] >= ' ') && (in[i] < 'z')) heading[t++] = in[i++];
			else i+=1;
		heading[t] = '\0';
		while ((in[i] != ',') && (i < size)) i += 1;

		//GET THE DATE TOKEN
        i++;
        t = 0;
        while ((in[i] != ',') && (in[i] != '.') && (t < 7) && (i < size))		
		    if ((in[i] >= ' ') && (in[i] < 'z')) date[t++] = in[i++];
			else i+=1;
		date[t] = '\0';

        convertCoordinate(latitude);
        convertCoordinate(longitude);

    } else {
        // bad checksum
        pushError(0x21);

    }// if


}// void parseDataRMC(char in[])


//-----------------------------------------------------------------------------
// CONVERTCOORDINATE
// Given: inCoordinate a pointer to coordinate "ddmm.mmmmm"
// Result: converts coordinate from "ddmm.ssss" format to "dd.dddddd"
//-----------------------------------------------------------------------------

void convertCoordinate(char coordinate[]) {
    int8 i = 0;
    char decimal[32];

    while ((coordinate[i] != '\0') && (coordinate[i] != '.') && (i < 12)) i++; // find index of decimal

    if ((i > 0) && (i < 12)) {
        strrmv(coordinate, i, i); // remove decimal from string
        ltoa(((atol(coordinate + (i - 2))*100) / 60), decimal);
        coordinate[i - 2] = '.';
        coordinate[i - 1] = '\0';
        strcat(coordinate, decimal);
    }// if
}// convertCoordinate(char *coordinate)


//-----------------------------------------------------------------------------
// CHECKSUM
// Given: {in} a pointer to a null terminated NMEA string
// Returns: TRUE is checksum is valid, FALSE if not
//-----------------------------------------------------------------------------
int8 checkSum(char in[]) {
    int8 computedSum = 0;
    int8 actualSum = 0;
    int8 i = 0;

    // compute checksum
    while ((in[i] != '*') &&(i<MAX_RMC_MESSAGE_LENGTH)) {
        if (in[i] != '$') computedSum ^= in[i];
        i++;
    }// while

	if (i == MAX_RMC_MESSAGE_LENGTH) return(FALSE);
 
    // get actual checksum
    i++;


    if ((in[i] >= '0') && (in[i] <= '9'))	actualSum += 16 * (in[i] - '0');
    else							        actualSum += 16 * (toupper(in[i]) - 'A' + 0x0A);
    if ((in[i+1] >= '0') && (in[i+1] <= '9'))	actualSum += (in[i+1] - '0');
    else									actualSum += (toupper(in[i+1]) - 'A' + 0x0A);

	// if (actualSum != computedSum) printf((far rom char *) "%x	%x\r\n",actualSum,computedSum);

    return (actualSum == computedSum);
}// checkSum(char *in)


//-----------------------------------------------------------------------------
// PUSHERROR
// Given: code ant 16 bit integer representing a specific error
// Result: code is pushed into the error queue
//-----------------------------------------------------------------------------

void pushError(int16 code) {
    INTCONbits.GIE = 0; //disable global interrupts

    if (++errorQueueSize <= 15) {
        errorQueue[errorQueueSize - 1] = code;
    } else {
        errorQueueSize = 1;
        errorQueue[errorQueueSize - 1] = code;
    }// if

    INTCONbits.GIE = 1; //enable global interrupts
}// pushError(int16 code)


//-----------------------------------------------------------------------------
// FRONTERROR
// Returns: value o item in the front of errorQueue (index zero)
//-----------------------------------------------------------------------------

int16 frontError(void) {
    int16 result = 0x00;
    if (errorQueueSize != 0)
        result = *errorQueue;

    return result;
}// frontError(void)


//-----------------------------------------------------------------------------
// BACKERROR
// Returns: value of item in the back of errorQueue (index (errorSize-1))
//-----------------------------------------------------------------------------

int16 backError(void) {
    int16 result = 0x00;
    if (errorQueueSize != 0)
        result = errorQueue[errorQueueSize - 1];

    return result;
}// backError(void)


//-----------------------------------------------------------------------------
// POPERROR
// Result: decrements errorQueueSize, removes front element and shifts all
//          elements up
//-----------------------------------------------------------------------------

void popError(void) {
    int8 i = 0;
    if (errorQueueSize > 0) {
        errorQueueSize--;

        while (i++ != errorQueueSize)
            errorQueue[i - 1] = errorQueue[i];
    }// if
}// popError(void)


//-----------------------------------------------------------------------------
// STRRMV
// Given:   s a pointer to a string
//          ini the initial index of the sub-string to be removed
//          fin the final index of the sub-string to be removed
// Result: Removes all character in {s} between {ini} and {fin}, and appends
//          a null terminator at the end of the string
//          example:    GIVEN   [s = "Daniel H.", ini = 3, fin = 5]
//                      RESULT  [s = "Dan H."]
//-----------------------------------------------------------------------------
void strrmv(char *s, int8 ini, int8 fin) {
    int8 size = strLength(s);
    fin++;
    while (fin < size) s[ini++] = s[fin++];
    s[ini] = '\0';
}// strrmv(char *s, int8 ini, int8 fin)


//-----------------------------------------------------------------------------
// STRLENGTH
// Given: a null terminated char array s
// Returns: number of characters preceding null character
//-----------------------------------------------------------------------------
int8 strLength(char s[]) {
    int8 i = 0;

    while ((s[i++] != '\0') && (i < MAX_STRING_LENGTH));
    return i-1;

}// strLength(char s[])

