//--------------------------------------------------------------------
//-- File:      trackABus.c (main)
//-- Name:		Daniel Hankewycz and Chris Coulston
//-- Date:		Spring/Summer 2013
//-- Purp: Bus Tracking firmware for the mobile unit on board the bus.
//			Responsible for reading the GPS NMEA strings and parsing-out
//			relevant information.  This information is tranmitted to the
//			server only when requsted by the server.
// 1:1024	Watchdog timer postscalar => 4 seconds
// 1:4096		=> 16 seconds
// 1:32768		=> 64 seconds
//-------------------------------------------------------------------------
#include <p18f26k22.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <usart.h>
#include <delays.h>
#include "myPic.h"
#include "GPS406a.h"

#define WDT_PRE	4096

/* Set up the configuration bits */
#pragma config FOSC = INTIO67
#pragma config PLLCFG = ON		// Multiply the internal oscillator speed by 4
#pragma config PRICLKEN = OFF	// Primary clock enabled
#pragma config FCMEN = OFF		// Fail-Safe Clock Monitor enabled
#pragma config IESO = OFF		// Oscillator Switchover mode disabled
#pragma config PWRTEN = OFF		// Power up timer disabled
#pragma config BOREN = OFF		// Brown-out Reset disabled in hardware and software
#pragma config BORV = 285		// VBOR set to 2.85 V nominal
#pragma config WDTEN = SWON		// WDT is controlled by SWDTEN bit of the WDTCON register .
#pragma config WDTPS = WDT_PRE		// Watchdog timer postscalar
#pragma config CCP2MX = PORTC1	// CCP2 input/output is multiplexed with RC1
#pragma config PBADEN = OFF		// PORTB<5:0> pins are configured as digital I/O on Reset
#pragma config CCP3MX = PORTB5	// P3A/CCP3 input/output is multiplexed with RB5
#pragma config HFOFST = OFF		// HFINTOSC output and ready status are delayed by the oscillator stable status
#pragma config T3CMX = PORTC0	// T3CKI is on RC0
#pragma config P2BMX = PORTC0	// P2B is on RC0
#pragma config MCLRE = EXTMCLR	// MCLR pin enabled, RE3 input pin disabled
#pragma config STVREN = OFF		// Stack full/underflow will not cause Reset
#pragma config LVP = OFF		// Single-Supply ICSP disabled
#pragma config XINST = OFF		// Instruction set extension and Indexed Addressing mode disabled (Legacy mode)
#pragma config DEBUG = ON		// Block 0 (000800-003FFFh) code-protected

//-------------------------------------------------------------------------
// Reserve space in the RAM for big data structures.  The program will 
// have two buffers for NMEA strings.  One of the buffers will be parsed
// while the other is being filled.  The buffers will alternate roles.
//-------------------------------------------------------------------------
#pragma udata redBuff
char priBuffer[100];
#pragma udata

#pragma udata blackBuff
char secBuffer[100];
#pragma udata

#pragma udata miscBuff
char longitude[12], latitude[12], heading[12], velocity[12], time[12], date[12], string[SCRATCH_STRING_LENGTH];
int8 gpsLock = FALSE;
int16 strNum = 0;
#pragma


int8 tour = 0;			// Which simulated GPS dataset is being fed to the system
int8 readyLine = FALSE;	// Whenever the GPS produces a new line of data
int8 buffReset = TRUE;	// You can reset the GPS buffer by setting this true
char *readFrom;			// Ptr to buffer which you read GPS NMEA string from
char *writeTo;			// Ptr to buffer the ISR is writing NMEA strings to
int8 gpsSimulation = FALSE; // Boolean to toggle source of NMEA strings
int8 usartPort;			// The destination of printf heads to the USART with this number


//----------------------------------------------
// This is the function prototype for the high
// priority ISR.
//----------------------------------------------
void high_isr(void);
#pragma code high_vector=0x08

void interrupt_at_high_vector(void) {
    _asm GOTO high_isr _endasm
}
#pragma code
void INIT_PIC(void);
void appendChecksum(char *string);

//----------------------------------------------
// Main "function"
//----------------------------------------------

void main(void) {

    int8 i, xor;

    readFrom = priBuffer;
    writeTo = secBuffer;
    latitude[0] = '\0';
    longitude[0] = '\0';
    time[0] = '\0';
    date[0] = '\0';
    velocity[0] = '\0';
    heading[0] = '\0';
	readFrom[0] = '\0';
	writeTo[0] = '\0';

    INIT_PIC();
	Delay10KTCYx (10);
    printf((far rom char *) "AVL rev 1.2\r\n");

	//----------------------------------------------
	// Configure the GPS module to run at 9600 baud
	// and transmitt only RMC strings once every 2
	// seconds.
	//----------------------------------------------
	SPBRG2 = 208;											// Slow down the USART to 4800 baud
	stdout = _H_USER;
	printf((far rom char *) "$PSRF103,00,00,00,01*24\r\n"); // Stop transmission of GGA
	Delay10KTCYx (10);
	printf((far rom char *) "$PSRF103,02,00,00,01*26\r\n"); // Stop transmission of GSA
	Delay10KTCYx (10);
	printf((far rom char *) "$PSRF103,04,00,02,01*22\r\n"); // RMC every 2 seconds
	Delay10KTCYx (10);
	printf((far rom char *) "$PSRF103,03,00,00,01*27\r\n"); // Stop transmission of GSV
	Delay10KTCYx (10);
	printf((far rom char *) "$PSRF100,1,9600,8,1,0*0D\r\n"); // 9600 baud
	Delay10KTCYx (10);
	stdout = _H_USART;
	SPBRG2 = 104;			// Resume operation at 9600 baud
	printf((far rom char *) "GPS module is now running @9600 baud.\r\n");


    while (TRUE) {

        if (TERM_FLAG) { // wait for the receive flag to be set
            TERM_FLAG = 0; // clear the flag for the next read
            GREEN_PORT = 0;
            switch (TERM_SRC) {

                    //----------------------------------------------
                    // Help menu for us humans
                    //----------------------------------------------
                case '?':
                    // printf((far rom char *) "\r\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
                    printf((far rom char *) "----------------- System Status --------\r\n");
                    printf((far rom char *) "	Tour		%d\r\n", tour);
                    printf((far rom char *) "	GPS src		");
                    if (gpsSimulation) printf((far rom char *) "simulation @ %d\r\n",strNum);
                    else printf((far rom char *) "USART2\r\n");
                    i = 0;
                    while (sim_name[i]) string[i++] = sim_name[i];
                    string[i] = 0x00;
                    printf((far rom char *) "	Sim type	%s\r\n", string);
                    printf((far rom char *) "	Output		USART%d\r\n", usartPort);
                    printf((far rom char *) "----------------- WDT configuration --------\r\n");
					if (WDTCONbits.SWDTEN == 1)	printf((far rom char *) "	WDT:		Enabled\r\n");
					else						printf((far rom char *) "	WDT:		Disabled\r\n");
					printf((far rom char *) "	WDT pre:	1:%d\r\n",WDT_PRE);
                    printf((far rom char *) "----------------- GPS Status -----------\r\n");
                    printf((far rom char *) "	Time		%s\r\n", time);
                    printf((far rom char *) "	Date		%s\r\n", date);
                    printf((far rom char *) "	Lat:Lon		%s:%s\r\n", latitude, longitude);
                    printf((far rom char *) "	Velocity	%s\r\n", velocity);
                    printf((far rom char *) "	Heading:	%s\r\n", heading);
					if (gpsLock == TRUE)	printf((far rom char *) "	GPS signal:	Locked\r\n");
					else					printf((far rom char *) "	GPS signal:	Scanning\r\n");
                    printf((far rom char *) "----------------- Query GPS ------------\r\n");
                    printf((far rom char *) "?: help		s: gps String	t: Time\r\n");
                    printf((far rom char *) "p: Position	l: Latitude	g: lonGitude\r\n");
                    printf((far rom char *) "v: Velocity	m: map		h: Heading\r\n");
                    printf((far rom char *) "i: GPS sIgnal status	d: Date\r\n");
                    printf((far rom char *) "--------------- Change system mode ------\r\n");
                    printf((far rom char *) "o: tOggle source of NMEA strings between simulation and USART2\r\n");
                    printf((far rom char *) "c: Change USART used by printf \r\n");
					printf((far rom char *) "k: Kick simuation ahead 30 seconds \r\n");
                    printf((far rom char *) "f: conFigure GPS to 9600 baud RMC\r\n");
                    printf((far rom char *) "F: configure GPS to Factory defaults\r\n");
                    printf((far rom char *) "R: Reboot the system\r\n"); 
                    printf((far rom char *) "-----------------------------------------\r\n");	
                    break;

                    //----------------------------------------------
                    // Raw NMA string - disable interrupts so that
                    // string is not corrupted mid-way through.
                    //----------------------------------------------
                case 's':
                    printf((far rom char *) "%s\r\n", readFrom);
                    break;

                    //----------------------------------------------
                    // Time string  hhmmss.sss
                    //----------------------------------------------
                case 't':
                    sprintf(string, (far rom char *) "$t %s#", time);
					appendChecksum(string);
					i=0; while(string[i]) putc(string[i++], _H_USART);
                    break;


                    //----------------------------------------------
                    // Date string Eurpoean Format ddmmyyyy
                    //----------------------------------------------
                case 'd':
                    sprintf(string, (far rom char *) "$d %s#", date);
					appendChecksum(string);
					i=0; while(string[i]) putc(string[i++], _H_USART);
                    break;


                    //----------------------------------------------
                    // Your position is has Lat and Long
                    //----------------------------------------------
                case 'p':
                    sprintf(string, (far rom char *) "$p %s ; %s#", latitude, longitude);
					appendChecksum(string);
					i=0; while(string[i]) putc(string[i++], _H_USART);
                    break;

                    //----------------------------------------------
                    // Latitude is in dd.mmmmmm format
                    //----------------------------------------------
                case 'l':
                    sprintf(string, (far rom char *) "$l %s#", latitude);
					appendChecksum(string);
					i=0; while(string[i]) putc(string[i++], _H_USART);
                    break;

                    //----------------------------------------------
                    // Longitute is in  dd.mmmmmm format
                    //----------------------------------------------
                case 'g':
                    sprintf(string, (far rom char *) "$g %s#", longitude);
					appendChecksum(string);
					i=0; while(string[i]) putc(string[i++], _H_USART);
                    break;

                    //----------------------------------------------
                    // The velocity is in ww.ff knots
                    //----------------------------------------------
                case 'v':
                    sprintf(string, (far rom char *) "$v %s#", velocity);
					appendChecksum(string);
					i=0; while(string[i]) putc(string[i++], _H_USART);
                    break;

                    //----------------------------------------------
                    // The data required to locate the bus on the map
                    //----------------------------------------------
                case 'm':
                    sprintf(string, (far rom char *) "$m %s ; %s ; %s#", latitude, longitude, heading);
					appendChecksum(string);
					i=0; while(string[i]) putc(string[i++], _H_USART);
                    break;

                    //----------------------------------------------
                    // Heading is in integer degrees with two fractional digits
                    //----------------------------------------------
                case 'h':
                    sprintf(string, (far rom char *) "$h %s#", heading);
					appendChecksum(string);
					i=0; while(string[i]) putc(string[i++], _H_USART);
                    break;


                    //----------------------------------------------
                    // Toggle the GPS acquisition mode between simulation and USART2
                    //----------------------------------------------
                case 'o':
                    if (gpsSimulation == TRUE) {
                        gpsSimulation = FALSE;
                        RCSTA2bits.CREN = 0;		// Clear the inevitable overrun error in RCREG2
                        printf((far rom char *) "NEMA strings coming from USART2\r\n");
                        T0CONbits.TMR0ON = 0;		// Turn off simulated NMEA strings from TMR0 interrupt
                        INTCONbits.TMR0IE = 0;		// and don't generate interrupts from this source
                        buffReset = TRUE;

                        RCSTA2bits.CREN = 1;		// and then reenable the receiver after the big printf delay
                        PIR3bits.RC2IF = 0;
                        PIE3bits.RC2IE = 1;

                    } else {
                        gpsSimulation = TRUE;
                        T0CONbits.TMR0ON = 1;
                        INTCONbits.TMR0IE = 1;
                        PIE3bits.RC2IE = 0;
                        printf((far rom char *) "NMEA strings provided by simluated data\r\n");
                    }
                    break;


                    //----------------------------------------------
                    // Toggle target of printf
                    //----------------------------------------------
                case 'c':
                    if (usartPort == 2) {
                        usartPort = 1;
                        stdout = _H_USART;
                    } else {
                        usartPort = 2;
                        stdout = _H_USER;
                    }

                    printf((far rom char *) "printf directed to USART%d\r\n", usartPort);
                    break;

                    //----------------------------------------------
                    // Kick simuatio forward 15 seconds
                    //----------------------------------------------
                case 'k':
					strNum += 30;
					printf((far rom char *) "Simulation now at: %d\r\n",strNum);
                    break;

                    //----------------------------------------------
                    // Return GPS signal status
					// A = Active
					// V = Void
                    //----------------------------------------------
                case 'i':
                    if (gpsLock == TRUE)	sprintf(string, (far rom char *) "$i A#");
                    else					sprintf(string, (far rom char *) "$i V#");
					appendChecksum(string);
					i=0; while(string[i]) putc(string[i++], _H_USART);
                    break;

                    //----------------------------------------------
                    // Configure the GPS module.  If doing this through
                    // hyperterminal, make sure to:
                    //	Edit -> Properties -> Settings tab -> Ascii Setup -> Send Line Ends with Line Feeds
                    //	then paste string using Right Mouse -> Paste to Host
                    // wasted an hour trying to figure that out :(
                    //
                    // Some useful commands
                    //	$PSRF103,00,00,00,01*24		Stop transmission of GGA
                    // 	$PSRF103,02,00,00,01*26		Stop transmission of GSA
                    // 	$PSRF103,03,00,00,01*27		Stop transmission of GSV
                    //	$PSRF103,04,00,02,01*22		RMC every 2 seconds
                    //	$PSRF100,1,9600,8,1,0*0D	9600 Baud
                    //
                    // Use the following for testing or restoring to factory defaults
                    //	$PSRF103,00,00,01,01*25		GGA every second
                    //	$PSRF103,02,00,01,01*27		GSA every second
                    //	$PSRF103,03,00,01,01*26		GSV every second
                    //	$PSRF103,04,00,00,01*20		Stop RMC
                    //	$PSRF100,1,4800,8,1,0*0E	4800 Baud
                    //----------------------------------------------
                case 'f':
                    printf((far rom char *) "Assumes GPS module was running on USART 2 @4800 baud.\r\n");
                    SPBRG2 = 208;			// Slow down the USART to 4800 baud
                    stdout = _H_USER;
                    printf((far rom char *) "$PSRF103,00,00,00,01*24\r\n"); // Stop transmission of GGA
					Delay10KTCYx (10);
                    printf((far rom char *) "$PSRF103,02,00,00,01*26\r\n"); // Stop transmission of GSA
					Delay10KTCYx (10);
                    printf((far rom char *) "$PSRF103,04,00,02,01*22\r\n"); // RMC every 2 seconds
					Delay10KTCYx (10);
                    printf((far rom char *) "$PSRF103,03,00,00,01*27\r\n"); // Stop transmission of GSV
					Delay10KTCYx (10);
                    printf((far rom char *) "$PSRF100,1,9600,8,1,0*0D\r\n"); // 9600 baud
                    Delay10KTCYx (10);
                    stdout = _H_USART;
                    SPBRG2 = 104;			// Resume operation at 9600 baud
                    printf((far rom char *) "GPS module is now running @9600 baud.\r\n");
                    break;
                    
                    //----------------------------------------------
                    // Configure the GPS module to factory defaults
                    //
                    // Use the following for testing or restoring to factory defaults
                    //	$PSRF103,00,00,01,01*25		GGA every second
                    //	$PSRF103,02,00,01,01*27		GSA every second
                    //	$PSRF103,03,00,01,01*26		GSV every second
                    //	$PSRF103,04,00,00,01*20		Stop RMC
                    //	$PSRF100,1,4800,8,1,0*0E	4800 Baud
                    //----------------------------------------------
                case 'F':
                    printf((far rom char *) "Assumes GPS module was running on USART 2 @9600 baud.\r\n");
                    stdout = _H_USER;
                    printf((far rom char *) "\n\r");
					Delay10KTCYx (10);
                    printf((far rom char *) "$PSRF103,00,00,01,01*25\r\n"); // GGA every second
					Delay10KTCYx (10);
                    printf((far rom char *) "$PSRF103,02,00,01,01*27\r\n"); // GSA every second
					Delay10KTCYx (10);
                    printf((far rom char *) "$PSRF103,03,00,01,01*26\r\n"); // GSV every second
					Delay10KTCYx (10);
                    printf((far rom char *) "$PSRF103,04,00,00,01*20\r\n"); // Stop RMC
					Delay10KTCYx (10);
                    printf((far rom char *) "$PSRF100,1,4800,8,1,0*0E\r\n"); // 4800 Baud
					Delay10KTCYx (10);
                    
                    if (usartPort == 2) stdout = _H_USER;
                    else stdout = _H_USART;
                    printf((far rom char *) "GPS module now running @4800 baud.\r\n");
                    break;

                    //----------------------------------------------
                    // The string "+++" allows the radio modem to
                    // enter command mode - don't mess this up with
					// the PIC chattering that it doesn't recognize
					// the "+" character
                    //----------------------------------------------
                case '+':
                    break;

                    //----------------------------------------------
                    // Raw NMA string - disable interrupts so that
                    // string is not corrupted mid-way through.
                    //----------------------------------------------
                case 'R':
                    _asm GOTO 0x0000 _endasm
                    break;

                    //----------------------------------------------
                    // Unknown/unsupported instruction
                    //----------------------------------------------
                default:
                    printf((far rom char *) "Unrecognized input: %c\r\n", TERM_SRC);
                    break;

            } // switch
            GREEN_PORT = 1;

		    //----------------------------------------------
			// Clear the Watch Dog Timer
 		   //----------------------------------------------
			_asm CLRWDT _endasm

        } // if

        //----------------------------------------------
        // If we have a new line then parse out the
        // relevant GPS data
        //----------------------------------------------
        if (readyLine) {
            if (getMode(readFrom) == GPRMC)
                parseDataRMC(readFrom);
            readyLine = FALSE;
        } // if

        //----------------------------------------------
        // Blue LED: simulation mode status
        //      ON:     SIM MODE
        //      OFF:    GPS MODE
        // Red LED: USART1 overflow status
        //      ON:     RX1 overrun error
        //      OFF:    NO overrun error
        //----------------------------------------------
        BLUE_PORT = !gpsSimulation;

        if (RCSTA1bits.OERR) {
            RED_PORT = 0;
			RCSTA1bits.CREN = 0;		// Clear the overrun 
			RCSTA1bits.CREN = 1;		// Clear the overrun 
        } else {
            RED_PORT = 1;
		}

        // Press select on joystick to clear the overrun error
        if((SELECT_PORT == 0) && (RCSTA1bits.OERR)) {
            RCSTAbits.CREN = 0;
            RCSTAbits.CREN = 1;
        }// if
    } // while forever
} // end main



//-----------------------------------------------------------------------------
// Interrupt handler for all the interrupts
//-----------------------------------------------------------------------------
#pragma interrupt high_isr

void high_isr(void) {

    static int8 buffIndex = 0;
    static int8 recording = FALSE;
    static int8 second = 0;
    int8 inChar, i;

    //----------------------------------------------
	// Clear the Watch Dog Timer
    //----------------------------------------------
	_asm CLRWDT _endasm

    //----------------------------------------------
    // TMR0 interrupt simulates the GPS
    //----------------------------------------------
    if (INTCONbits.TMR0IF && !gpsSimulation) {
        LATCbits.LATC1 ^= 1;
        INTCONbits.TMR0IF = 0; // Clear the flag
    }

    if (INTCONbits.TMR0IF && gpsSimulation) {
        INTCONbits.TMR0IF = 0; // Clear the flag
        if (second++ == 1) {
            // Crude string copy from program memory to data memory
            i = 0;
            while (sim[strNum][i]) writeTo[i++] = sim[strNum][i];
            writeTo[i] = 0x00;

            if (writeTo == secBuffer) {
                writeTo = priBuffer;
                readFrom = secBuffer;
            } else {
                writeTo = secBuffer;
                readFrom = priBuffer;
            } // end if (buffer pointer swap)

            if (strNum >= sim_length) strNum = 0;
            else strNum += 1;
            readyLine = TRUE;
            second = 0;
        } // end 2 second interval
        TMR0H = 0x0B;
        TMR0L = 0xDD;
    } // end


    //----------------------------------------------
    // The GPS module sends characters on USART2
    //----------------------------------------------
    if (PIR3bits.RC2IF && !gpsSimulation) {

        PIR3bits.RC2IF = 0; // Make sure to clear the flag
        inChar = RCREG2; //

        //----------------------------------------------
        // You can tell the ISR to ignore the rest of
        // the current string and look the start of a
        // new string "$" by setting buffRest = TRUE
        //----------------------------------------------
        if (buffReset) {
            buffReset = FALSE;
            recording = FALSE;
        }

        //----------------------------------------------
        // Only 3 cases are relevant:
        //		$	=> Start of NMEA string
        //		\r	=> End of NMEA string
        //			=> A character to add to the string
        //----------------------------------------------
        switch (inChar) {

            case '$':
                buffIndex = 0;
                writeTo[buffIndex++] = '$';
                recording = TRUE;
                break;

            case '\r':
                writeTo[buffIndex] = 0x00;
                if (writeTo == secBuffer) {
                    writeTo = priBuffer;
                    readFrom = secBuffer;
                } else {
                    writeTo = secBuffer;
                    readFrom = priBuffer;
                } // end if (buffer pointer swap)
                readyLine = TRUE;
                recording = FALSE;
                break;

            default:
                if (recording) writeTo[buffIndex++] = inChar;
                break;
        } // end character received
    } // end USART 2 interrupt
} // high_isr()


//----------------------------------------------
// INIT_PIC
// Initializes the PIC to run at 64MHz
// Setup a terminal interface 9600 on COM1 and COM2
//----------------------------------------------

void INIT_PIC(void) {

    // ---------------Configure Oscillator------------------
    OSCCONbits.IRCF2 = 1; // Internal RC Oscillator Frequency Select bits
    OSCCONbits.IRCF1 = 1; // Set to 16Mhz
    OSCCONbits.IRCF0 = 1; //
    OSCTUNEbits.PLLEN = 1; // enable the 4xPLL, wicked fast 64Mhz

    //-----------Configure timer0 to interrupts every 2 seconds----------------
    // At 3.814hz
    T0CON = 0; 				// Funky power-on defaults, see page 159
    T0CONbits.PSA = 0;		// Assign prescalar to TMR0
    T0CONbits.T0PS2 = 1;	// 001 = 1:4		101 = 1:64
    T0CONbits.T0PS1 = 1;	// 010 = 1:8		110 = 1:128
    T0CONbits.T0PS0 = 1;	// 011 = 1:16		111 = 1:256
    T0CONbits.T08BIT = 0;	// 8-bit mode is legacy for older devices
    T0CONbits.TMR0ON = 0;	// best to configure devices then turn them on
    TMR0H = 0x00;
    TMR0L = 0x00;
    INTCONbits.TMR0IF = 0; // Clear interrupt flag

    // ---------------Setup the serial port------------------
    // Setup both terminals to operate at 9600 Baud
    // BAUD = FOSC/[64*(SPBRGH+1)]
    // SPBRGH = 64Mhz/ 9.6k/64 =  104

    TXSTA1bits.TXEN = 1;
    TXSTA1bits.SYNC = 0;
    TXSTA1bits.BRGH = 0;
    BAUDCON1bits.BRG16 = 0;
    RCSTA1bits.CREN = 1;
    SPBRG1 = 104;
    RCSTA1bits.SPEN = 1;
    TRISCbits.TRISC7 = INPUT;
    ANSELCbits.ANSC7 = DIGITAL;

    TXSTA2bits.TXEN = 1;
    TXSTA2bits.SYNC = 0;
    TXSTA2bits.BRGH = 0;
    BAUDCON2bits.BRG16 = 0;
    RCSTA2bits.CREN = 1;
    SPBRG2 = 104;
    RCSTA2bits.SPEN = 1;
    TRISBbits.TRISB7 = INPUT;

    //-------------------CONFIGURE RGB LED-------------------
    // Leds are active low
    // RED:     RA3
    // GREEN:   RA4
    // BLUE:    RA5
    ANSELAbits.ANSA3 = DIGITAL;
    ANSELAbits.ANSA5 = DIGITAL;
    TRIS_RED_PIN = OUTPUT;
    TRIS_GREEN_PIN = OUTPUT;
    TRIS_BLUE_PIN = OUTPUT;
    RED_PORT = 1;
    GREEN_PORT = 1;
    BLUE_PORT = 1;

    //-------------------CONFIGURE JOYSTICK-------------------
    // Buttons are common ground
    // UP:      B3
    // DOWN:    B4
    // LEFT:    B0
    // RIGHT:   B1
    // SELECT:  B2
    ANSELBbits.ANSB0 = DIGITAL;
    ANSELBbits.ANSB1 = DIGITAL;
    ANSELBbits.ANSB2 = DIGITAL;
    ANSELBbits.ANSB3 = DIGITAL;
    ANSELBbits.ANSB4 = DIGITAL;
    TRIS_UP_PIN = INPUT;
    TRIS_DOWN_PIN = INPUT;
    TRIS_LEFT_PIN = INPUT;
    TRIS_RIGHT_PIN = INPUT;
    TRIS_SELECT_PIN = INPUT;

    // The user can decide which USART receives output (via printf) by the usartPORT variable
    if (USART1 == TRUE) { // This is a define in myPic.h
        usartPort = 1; // The destination of printf heads to the USART with this number
        stdout = _H_USART; // Tells printf to use (default) USART1
    } else {
        usartPort = 2;
        stdout = _H_USER; // Tells printf to use the _user_putc function (route to USART #2)
    }
    buffReset = TRUE;
    //-----------------Setup interrupts--------------------
    if (gpsSimulation == TRUE) {
        INTCONbits.TMR0IE = 1; // Enable TMR0 interrupt
        T0CONbits.TMR0ON = 1;
    } else PIE3bits.RC2IE = 1; // enables PORT 2 RX interrupt
    INTCONbits.PEIE = 1; // enables peripheral interrupts
    INTCONbits.GIE = 1; // enables global interrupts

    //-----------------For testing I/O configurations--------------------
    TRISCbits.TRISC1 = OUTPUT;
    TRISBbits.TRISB0 = INPUT;
    ANSELBbits.ANSB0 = DIGITAL;
    INTCON2bits.RBPU = 0;
    WPUBbits.WPUB0 = 1;
    WPUBbits.WPUB7 = 0;

    //-----------------Watch Dog Timer configuration--------------------
	WDTCONbits.SWDTEN = 1;			// Turn off the WDT - return to this at some later point in time

} // end initialize the PIC



//-----------------------------------------------------------------------------
// Helper function needed to point PRINTF to the second USART
//-----------------------------------------------------------------------------

int _user_putc(char c) {

    while (TXSTA2bits.TRMT != 1); // The TRMT bit is set when the TSR register is empty
    TXREG2 = c;

}


//-----------------------------------------------------------------------------
// Tags a string with a checksum
//-----------------------------------------------------------------------------
void appendChecksum(char *string) {

	int8 i, xor, nybble;

	i = 1;
	xor = 0;
	while ((string[i] != '#')&&(i<SCRATCH_STRING_LENGTH)) xor ^= string[i++];

	nybble = (xor&0xF0) >>4;
	if (nybble <= 9) string[i+1] = '0' + nybble; else string[i+1] = 'a' + nybble - 10;

	nybble = xor&0x0F;
	if (nybble <= 9) string[i+2] = '0' + nybble; else string[i+2] = 'a' + nybble - 10;

	string[i+3] = '\r';
	string[i+4] = '\n';
	string[i+5] = 0x00;

}