// Original file header: 
// NAA - NOAA's Astronomical Algorithms
// (JavaScript web page http://www.srrb.noaa.gov/highlights/sunrise/sunrise.html by
// Chris Cornwall, Aaron Horiuchi and Chris Lehman)
// Ported to C# by Pete Gray (petegray@ieee.org), June 2006
// Released as Open Source and can be used in any way, as long as the above description
// remains in place.
// Ported to C by S.S. Brandsma (sbrandsma@gmail.com) for the Renesas Green Challelange project, july 2012

/*
*********************************************************************************************************
*
*                                          AstroMath
*                                     
*                                          Contains all functions to go from 
*                                          Lon Lat position to position of the sun
*                                          With that information, heading/yaw and pitch
*                                          can be calculated
*
* Filename      : astro_math.c
* Version       : V1.00
*********************************************************************************************************
*/

/*
*********************************************************************************************************
*                                             INCLUDE FILES
*********************************************************************************************************
*/
#include <includes.h>
#include <stdint.h>
#include <CG_macrodriver.h> 
#include <CG_userdefine.h>
#include <monitor.h>


/*
*********************************************************************************************************
*                                             DEFINES
*********************************************************************************************************
*/



/*
*********************************************************************************************************
*                                            Function protoypes
*********************************************************************************************************
*/
void ApproximateSolarCoordinates(DateTime *timeStamp, float lat, float lon, float *azimuth, float *elevation);
float ApproximateSolarElevation(DateTime *timeStamp, float lat, float lon);
float JD(DateTime *timeStamp);       
void EquatorialToHorizontal(DateTime *timeStamp, float lat, float lon, float RA, float dec, float *az, float *el);
float MeanSiderealTime(DateTime *timeStamp, float longitude);
float calcJD(int year, int month, int day);  
float calcTimeJulianCent(float jd);
float calcJDFromJulianCent(float t);
float calcGeomMeanLongSun(float t);
float calcGeomMeanAnomalySun(float t);
float calcEccentricityEarthOrbit(float t);
float calcSunEqOfCenter(float t);
float calcSunTrueLong(float t);
float calcSunTrueAnomaly(float t);
float calcSunRadVector(float t);
float calcSunApparentLong(float t);
float calcMeanObliquityOfEcliptic(float t);
float calcObliquityCorrection(float t);
float calcSunRtAscension(float t);
float calcSunDeclination(float t);
float calcEquationOfTime(float t);
float calcHourAngleSunrise(float lat, float solarDec);
float calcHourAngleSunset(float lat, float solarDec);

  
// <summary>
// Computes approximate solar coordinates wrt the local horizontal plane
// </summary>
// <param name="timeStamp">Time (UTC)</param>
// <param name="lat">Latitude of observer [rad]</param>
// <param name="lon">Longitude of observer [rad]</param>
// <param name="azimuth">Azimuth to the sun [rad]</param>
// <param name="elevation">Elevation to the sun [rad]</param>
void ApproximateSolarCoordinates(DateTime *timeStamp, float lat, float lon, float *azimuth, float *elevation)
{
  float D  = JD(timeStamp) - 2451545.0;
  float g  = DEG2RAD(357.529 + 0.98560028 * D);
  float q  = DEG2RAD(280.459 + 0.98564736 * D);
  float L  = q + DEG2RAD(1.915 * sin(g) + 0.020 * sin(2.0 * g));
  float e  = DEG2RAD(23.439 - 0.00000036 * D);
  float RA = atan2(cos(e) * sin(L), cos(L));
  float d  = asin(sin(e) * sin(L));

  EquatorialToHorizontal(timeStamp, lat, lon, RA, d, azimuth, elevation);
}

// <summary>
// Computes approximate solar elevation wrt the local horizontal plane
// </summary>
// <param name="timeStamp">Time (UTC)</param>
// <param name="lat">Latitude of observer [rad]</param>
// <param name="lon">Longitude of observer [rad]</param>
// <returns>Approximate solar elevation</returns>
float ApproximateSolarElevation(DateTime *timeStamp, float lat, float lon)
{
  float D = JD(timeStamp) - 2451545.0;
  float g = DEG2RAD(357.529 + 0.98560028 * D);
  float q = DEG2RAD(280.459 + 0.98564736 * D);
  float L = q + DEG2RAD(1.915 * sin(g) + 0.020 * sin(2.0 * g));
  float e = DEG2RAD(23.439 - 0.00000036 * D);
  float RA = atan2(cos(e) * sin(L), cos(L));
  float d = asin(sin(e) * sin(L));

  float azimuth   = 0.0;
  float elevation = 0.0;
  EquatorialToHorizontal(timeStamp, lat, lon, RA, d, &azimuth, &elevation);
  return elevation;
}

// <summary>
// Returns Julian Date (JD)
// </summary>
float JD(DateTime *timeStamp)       
{
  float hour = timeStamp->Hour + (timeStamp->Min / 60.0) + (timeStamp->Sec / 3600.0);
  float GGG  = 1.0; //if (timeStamp->Year <= 1585) GGG = 0;

  volatile float JD   = -1 * floor(7.0 * (floor((timeStamp->Month + 9.0) / 12.0) + timeStamp->Year) / 4.0);
  float S    = 1.0; if ((timeStamp->Month - 9) < 0) S = -1;

  float A    = abs(timeStamp->Month - 9);
  float J1   = floor(timeStamp->Year + S * floor(A / 7.0));

  J1       = -1 * floor((floor(J1 / 100) + 1) * 3 / 4);
  JD       += floor(275.0 * timeStamp->Month / 9.0) + timeStamp->Day + (GGG * J1);
  JD       += 1721027.0 + 2.0 * GGG + 367.0 * timeStamp->Year - 0.5;
  JD       += (hour / 24.0);
  return JD;
}

// <summary>
// Convert equatorial coordinates to horizontal
// </summary>
// <param name="timeStamp">Timestamp</param>
// <param name="lat">Latitude of observer [rad]</param>
// <param name="lon">Longitude of observer [rad]</param>
// <param name="RA">Right ascension [rad]</param>
// <param name="dec">Declination [rad]</param>
// <param name="az">Azimuth [rad]</param>
// <param name="el">Elevation [rad]</param>

void EquatorialToHorizontal(DateTime *timeStamp, float lat, float lon, float RA, float dec, float *az, float *el)
{
  float ha = MeanSiderealTime(timeStamp, lon) - RA; // Hour Angle [radians]
  if (ha < DOUBLE_PI) ha += DOUBLE_PI;

  *el = asin(sin(dec) * sin(lat) + cos(dec) * cos(lat) * cos(ha));
  *az = acos((sin(dec) - sin(*el) * sin(lat)) / (cos(*el) * cos(lat)));

  if (sin(ha) > 0.0) 
    (*az) = DOUBLE_PI - (*az);
}

// <summary>
// Returns Mean Sidereal Time in radians
// http://home.att.net/~rschmitt/celestial2horizon.html
// </summary>
// <param name="timeStamp">Local time</param>
// <param name="longitude">Local longitude [radians]</param>
// <returns>Time [radians]</returns>
float MeanSiderealTime(DateTime *timeStamp, float longitude)
{
  unsigned short year   = timeStamp->Year;
  uint8_t month  = timeStamp->Month;
  uint8_t day    = timeStamp->Day;
  uint8_t hour   = timeStamp->Hour;
  uint8_t Min    = timeStamp->Min;
  uint8_t Sec    = timeStamp->Sec;

  if (month == 1 | month == 2)
  {
    year--;
    month += 12;
  }

  float a = floor((float)year / 100);
  float b = 2 - a + floor(a / 4);
  float c = floor(365.25 * year);
  float d = floor(30.6001 * (month + 1));

  float jd  = b + c + d - 730550.5 + day + (hour + Min / 60.0 + Sec / 3600.0) / 24.0;
  float jt  = jd / 36525.0;
  float mst = DEG2RAD(280.46061837 + 360.98564736629 * jd + 0.000387933 * jt * jt - jt * jt * jt / 38710000.0) + longitude;
  if (mst > 0.0)
  {
   // while (mst > 2 * PI)
   // {
   //   mst -= 2 * PI;
   // }
      mst = mst - ((long) (mst / (DOUBLE_PI))) * (DOUBLE_PI);
  }
  else
  {
   // while (mst < 2 * PI)
   // {
   //   mst += 2 * PI;
   // }
      mst = mst - ((long) (mst / (DOUBLE_PI))) * (DOUBLE_PI) + (DOUBLE_PI);
  }
  return mst;
}

// <summary>
// Julian day from calendar day 
// </summary>
// <remarks>
// Number is returned for start of day. Fractional days should be added later
// </remarks>
// <param name="year"></param>
// <param name="month"></param>
// <param name="day"></param>
// <returns></returns> 
float calcJD(int year, int month, int day)
{
  if (month <= 2)
  {
    year  -= 1;
    month += 12;
  }

  float A = floor(year / 100.0);
  float B = 2 - A + floor(A / 4);

  return (floor(365.25 * (year + 4716)) + floor(30.6001 * (month + 1)) + day + B - 1524.5);
}

// <summary>
// Convert Julian Day to centuries since J2000.0.
// </summary>
// <param name="jd">The Julian Day to convert</param>
// <returns>value corresponding to the Julian Day</returns>
float calcTimeJulianCent(float jd)
{
  return ((jd - 2451545.0) / 36525.0);
}

// <summary>
// Convert centuries since J2000.0 to Julian Day
// </summary>
// <param name="t">Number of Julian centuries since J2000.0</param>
// <returns>Julian Day corresponding to the t value</returns>
float calcJDFromJulianCent(float t)
{
  return (t * 36525.0 + 2451545.0);
}

// <summary>
// Calculate the Geometric Mean Longitude of the Sun
// </summary>
// <param name="t">number of Julian centuries since J2000.0</param>
// <returns>the Geometric Mean Longitude of the Sun in degrees</returns>
float calcGeomMeanLongSun(float t)
{
  float L0 = 280.46646 + t * (36000.76983 + 0.0003032 * t);
  while (L0 > 360.0)
  {
    L0 -= 360.0;
  }
  while (L0 < 0.0)
  {
    L0 += 360.0;
  }
  return L0; // in degrees
}

// <summary>
// Calculate the Geometric Mean Anomaly of the Sun
// </summary>
// <param name="t">Number of Julian centuries since J2000.0</param>
// <returns></returns>
float calcGeomMeanAnomalySun(float t)
{
  return (357.52911 + t * (35999.05029 - 0.0001537 * t));
}

// <summary>
// Calculate the eccentricity of earth's orbit
// </summary>
// <param name="t">Number of Julian centuries since J2000.0</param>
// <returns>Eccentricity (unitless)</returns>
float calcEccentricityEarthOrbit(float t)
{
  return (0.016708634 - t * (0.000042037 + 0.0000001267 * t));
}

// <summary>
// Calculate the equation of center for the sun [deg]
// </summary>
// <param name="t">number of Julian centuries since J2000.0</param>
// <returns></returns>
float calcSunEqOfCenter(float t)
{
  float m     = calcGeomMeanAnomalySun(t);

  float mrad  = DEG2RAD(m);
  float sinm  = sin(mrad);
  float sin2m = sin(mrad + mrad);
  float sin3m = sin(mrad + mrad + mrad);

  return sinm * (1.914602 - t * (0.004817 + 0.000014 * t)) + sin2m * (0.019993 - 0.000101 * t) + sin3m * 0.000289; // in degrees
}

// <summary>
// Calculate the true longitude of the sun
// </summary>
// <param name="t">Number of Julian centuries since J2000.0</param>
// <returns>Sun's true longitude in degrees</returns>
float calcSunTrueLong(float t)
{
  float l0 = calcGeomMeanLongSun(t);
  float c  = calcSunEqOfCenter(t);

  return l0 + c; // In degrees
}

// <summary>
// Calculate the true anamoly of the sun
// </summary>
// <param name="t">Number of Julian centuries since J2000.0</param>
// <returns>Sun's true anamoly in degrees</returns>
float calcSunTrueAnomaly(float t)
{
  float m = calcGeomMeanAnomalySun(t);
  float c = calcSunEqOfCenter(t);

  return m + c; // In degrees
}

// <summary>
// Calculate the distance to the sun in AU
// </summary>
// <param name="t">Number of Julian centuries since J2000.0</param>
// <returns>Sun radius vector in AUs</returns>
float calcSunRadVector(float t)
{
  float v = calcSunTrueAnomaly(t);
  float e = calcEccentricityEarthOrbit(t);

  return (1.000001018 * (1 - e * e)) / (1 + e * cos(DEG2RAD(v))); // In AU's
}

// <summary>
// Calculate the apparent longitude of the sun
// </summary>
// <param name="t">Number of Julian centuries since J2000.0</param>
// <returns>Sun's apparent longitude in degrees</returns>
float calcSunApparentLong(float t)
{
  float o     = calcSunTrueLong(t);
  float omega = 125.04 - 1934.136 * t;

  return o - 0.00569 - 0.00478 * sin(DEG2RAD(omega)); // In Degrees
}

// <summary>
// Calculate the mean obliquity of the ecliptic
// </summary>
// <param name="t">Number of Julian centuries since J2000.0</param>
// <returns>Mean obliquity in degrees</returns>
float calcMeanObliquityOfEcliptic(float t)
{
  float Secs = 21.448 - t * (46.8150 + t * (0.00059 - t * (0.001813)));
  return 23.0 + (26.0 + (Secs / 60.0)) / 60.0; // In degrees
}

// <summary>
// Calculate the corrected obliquity of the ecliptic 
// </summary>
// <param name="t">Number of Julian centuries since J2000.0</param>
// <returns>Corrected obliquity in degrees</returns>
float calcObliquityCorrection(float t)
{
  float e0    = calcMeanObliquityOfEcliptic(t);
  float omega = 125.04 - 1934.136 * t;

  return e0 + 0.00256 * cos(DEG2RAD(omega)); // In degrees
}

// <summary>
// Calculate the right ascension of the sun
// </summary>
// <param name="t">Number of Julian centuries since J2000.0</param>
// <returns>Sun's right ascension in degrees</returns>
float calcSunRtAscension(float t)
{
  float e         = calcObliquityCorrection(t);
  float lambda    = calcSunApparentLong(t);
  float tananum   = (cos(DEG2RAD(e)) * sin(DEG2RAD(lambda)));
  float tanadenom = (cos(DEG2RAD(lambda)));

  return RAD2DEG(atan2(tananum, tanadenom));
}

// <summary>
// Calculate the declination of the sun
// </summary>
// <param name="t">Number of Julian centuries since J2000.0</param>
// <returns>Sun's declination in degrees</returns>
float calcSunDeclination(float t)
{
  float e      = calcObliquityCorrection(t);
  float lambda = calcSunApparentLong(t);
  float sint   = sin(DEG2RAD(e)) * sin(DEG2RAD(lambda));

  return RAD2DEG(asin(sint));
}

// <summary>
// Calculate the difference between true solar time and mean solar time
// </summary>
// <param name="t">number of Julian centuries since J2000.0</param>
// <returns>Equation of time in Mins of time</returns>
float calcEquationOfTime(float t)
{
  float epsilon = calcObliquityCorrection(t);
  float l0      = calcGeomMeanLongSun(t);
  float e       = calcEccentricityEarthOrbit(t);
  float m       = calcGeomMeanAnomalySun(t);

  float y = tan(DEG2RAD(epsilon) / 2.0);
  y *= y;

  float sin2l0 = sin(2.0 * DEG2RAD(l0));
  float sinm   = sin(DEG2RAD(m));
  float cos2l0 = cos(2.0 * DEG2RAD(l0));
  float sin4l0 = sin(4.0 * DEG2RAD(l0));
  float sin2m  = sin(2.0 * DEG2RAD(m));
  float Etime  = y * sin2l0 - 2.0 * e * sinm + 4.0 * e * y * sinm * cos2l0 - 0.5 * y * y * sin4l0 - 1.25 * e * e * sin2m;

  return RAD2DEG(Etime) * 4.0; // in Mins of time
}

// <summary>
// Calculate the hour angle of the sun at sunrise for the latitude
// </summary>
// <param name="lat">Latitude of observer in degrees</param>
// <param name="solarDec">Declination angle of sun in degrees</param>
// <returns>Hour angle of sunrise in radians</returns>
float calcHourAngleSunrise(float lat, float solarDec)
{
  float latRad = DEG2RAD(lat);
  float sdRad  = DEG2RAD(solarDec);

  return (acos(cos(DEG2RAD(90.833)) / (cos(latRad) * cos(sdRad)) - tan(latRad) * tan(sdRad))); // In radians
}

// <summary>
// Calculate the hour angle of the sun at sunset for the latitude
// </summary>
// <param name="lat">Latitude of observer in degrees</param>
// <param name="solarDec">Declination angle of sun in degrees</param>
// <returns>Hour angle of sunset in radians</returns>
float calcHourAngleSunset(float lat, float solarDec)
{
  float latRad = DEG2RAD(lat);
  float sdRad  = DEG2RAD(solarDec);

  return -(acos(cos(DEG2RAD(90.833)) / (cos(latRad) * cos(sdRad)) - tan(latRad) * tan(sdRad))); // In radians
}
