/*
*********************************************************************************************************
*
*                                        RDKRL78 SPI driver
*                                     
*                                           Renesas RL78
*                                               on the
*                                     YRDKRL78G13 Evaluation Board
*
* Filename      : rdkrl78_spi.c
* Version       : V1.00
*********************************************************************************************************
*/


/*
*******************************************************************************
** Include files
*******************************************************************************
*/
#include <includes.h>
#include <CG_macrodriver.h>
#include <CG_userdefine.h>
#include <CG_serial.h>
#include <CG_it.h>
#include <RDKRL78_spi.h>

/*
*******************************************************************************
**  Local define
*******************************************************************************
*/
// SD Card
#define SD_SEND_STATUS 0x13
#define SD_GO_IDLE_CMD	0x40
#define SD_GO_IDLE_CRC	0x95

/*
*********************************************************************************************************
*                                         FUNCTION PROTOTYPES
*********************************************************************************************************
*/
void IO_Reset();

/*
*********************************************************************************************************
**                                            GLOBALS / EXTERNALS
*********************************************************************************************************
*/
extern volatile UCHAR G_SPI_SendingData; //spi busy flag
extern volatile UCHAR G_SPI_ReceivingData; //spi busy flag

/*
*********************************************************************************************************
**                                            LOCALS
*********************************************************************************************************
*/
static bool initDone = false;

unsigned char *SPI_CS_Port[] = {
    (unsigned char *)&P1, // LCD-CS    P10
    (unsigned char *)&P3, // SD-CS     P30
    (unsigned char *)&P7, // PMOD1-CS  P71
    (unsigned char *)&P7  // PMOD2-CS  P72
};

uint8_t SPI_CS_Pin[] = {
    0, // LCD-CS    P10
    0, // SD-CS     P30
    1, // PMOD1-CS  P71
    2  // PMOD2-CS  P72
};




/*
*********************************************************************************************************
*                                          SPI_Init
*
* Description : Initializes the SPI interface
*
* Arguments   : none
*
* Returns     : none
*
* Notes       : none
*********************************************************************************************************
*/
void SPI_Init()
{   
   if (initDone)
     return;

    IO_Reset();    
    SAU0_Init();

    CSI10_Start();
}


/*
*********************************************************************************************************
*                                          IO_Reset
*
* Description : Resets SPI
*
* Arguments   : none
*
* Returns     : none
*
* Notes       : none
*********************************************************************************************************
*/
void IO_Reset()
{
    int i = 0;
    
    //#warning RESET-IO must be inverted for actual HW
    P13 |= (1<<0); // Assert P130 (#RESET-IO)
    for (i=0;i<10000;i++);
    P13 &= ~(1<<0);  // Deassert P130 (#RESET-IO)
    for (i=0;i<10000;i++);
}


/*
*********************************************************************************************************
*                                          SPI_CS_Start
*
* Description : Enables SPI for the specified device
*
* Arguments   : none
*
* Returns     : none
*
* Notes       : none
*********************************************************************************************************
*/
void SPI_CS_Start(uint8_t aDevice)
{
	*SPI_CS_Port[aDevice] &= ~(1<<SPI_CS_Pin[aDevice]);

}


/*
*********************************************************************************************************
*                                          SPI_CS_End
*
* Description : Disables SPI for the specified device
*
* Arguments   : none
*
* Returns     : none
*
* Notes       : none
*********************************************************************************************************
*/
void SPI_CS_End(uint8_t aDevice)
{
	*SPI_CS_Port[aDevice] |= (1<<SPI_CS_Pin[aDevice]);
}

/*
*********************************************************************************************************
*                                          SPI_Send
*
* Description : Sends data over the SPI bus 
*
* Arguments   : aDevice: Device as specified at the SPI_CS_Portobject
*               aData: pointer to data to be transmitted
*               aLength: length of data to be transmitted
*
* Returns     : none
*
* Notes       : This is a blocking call
*********************************************************************************************************
*/
void SPI_Send(uint8_t aDevice, uint8_t *aData, uint32_t aLength)
{
    uint8_t noRXData;
    G_SPI_SendingData = 1;
    G_SPI_ReceivingData = 0;
	
    SPI_CS_Start(aDevice);

    CSI10_SendReceiveData(aData, aLength, &noRXData,0);
    while(G_SPI_SendingData);
    
    SPI_CS_End(aDevice);
}


/*
*********************************************************************************************************
*                                          SPI_Receive
*
* Description : Receives data from the SPI bus
*
* Arguments   : aDevice: Device as specified at the SPI_CS_Portobject
*               aData: pointer to data to be received
*               aLength: Number of bytes to be expected back
*
* Returns     : none
*
* Notes       : This is a blocking call
*********************************************************************************************************
*/
void SPI_Receive(uint8_t aDevice, uint8_t *aData, uint32_t aLength)
{
    uint8_t noTXData = 0xFF;
    G_SPI_SendingData = 0;
    G_SPI_ReceivingData = 1;
	
    SPI_CS_Start(aDevice);

    CSI10_SendReceiveData(&noTXData, 1, aData, aLength );
    while(G_SPI_ReceivingData);
    
    SPI_CS_End(aDevice);
}

/*
*********************************************************************************************************
*                                          SPI_SendReceive
*
* Description : Sends data over the SPI bus and expects result back based on the the given length
*
* Arguments   : aDevice: Device as specified at the SPI_CS_Portobject
*               aTXData: pointer to data to be transmitted
*               aTXLength: length of data to be transmitted
*               aRXData: pointer to data to be received
*               aRXLength: Number of bytes to be expected back
*
* Returns     : none
*
* Notes       : This is a blocking call
*********************************************************************************************************
*/
void SPI_SendReceive(uint8_t aDevice, uint8_t *aTXData, uint32_t aTXLength, uint8_t *aRXData, uint32_t aRXLength )
{
    G_SPI_SendingData = 1;
    G_SPI_ReceivingData = 1;
    
    SPI_CS_Start(aDevice);
	
    delay_ms(1);
    
    CSI10_SendReceiveData(aTXData, aTXLength, aRXData, aRXLength);
    while(G_SPI_SendingData || G_SPI_ReceivingData);
    
	SPI_CS_End(aDevice);
}

