/*
*********************************************************************************************************
*
*                                           Controller Task
*                                     
*                                           Renesas RL78
*                                               on the
*                                     YRDKRL78G13 Evaluation Board
*
* Filename      : controller.c
* Version       : V1.00
*********************************************************************************************************
*/

/*
*********************************************************************************************************
*                                             INCLUDE FILES
*********************************************************************************************************
*/
#include <includes.h>
#include <string.h>
#include <CG_macrodriver.h>
#include <CG_serial.h>
#include <CG_timer.h>
#include <CG_userdefine.h>
#include <ff.h>
#include <RDKRL78_spi.h>
#include <controller.h>
#include <monitor.h>

/*
*********************************************************************************************************
*                                             LOCAL DEFINES
*********************************************************************************************************
*/



/*
*********************************************************************************************************
**                                            GLOBALS / EXTERNALS
*********************************************************************************************************
*/
OS_Q             ControlQueue;
extern DateTime dateTime;

/*
*********************************************************************************************************
*                                            LOCAL VARIABLES
*********************************************************************************************************
*/
static  OS_ERR    err;
static  OS_TCB    App_ControlTaskTCB; 
static  CPU_STK   App_ControlTaskStk[APP_CFG_TASK_CONTROL_STK_SIZE];

static  const UCHAR steps[] = {STEP1,STEP2,STEP3,STEP4};
static  UCHAR stepperPosition = 0;

static  FATFS fs;                                   /* Work area (file system object) for logical drives */
static  FIL fsrc;                                   /* file objects */
static  FRESULT res = FR_INVALID_PARAMETER;         /* FatFs function common result code */

/*
*********************************************************************************************************
*                                         FUNCTION PROTOTYPES
*********************************************************************************************************
*/
static  void  App_TaskCreate(void);
static  void  App_ObjCreate (void);
static  void  App_ControlTask (void  *p_arg);

static  void rotatePlatform();                    /* Rotates the platform */
static  void stepUp();                            /* tilt platform up */
static  void stepDown();                          /* tilt platform down */
static  void FS_Init();                           /* Initialize FAT16/32 filesystem */

/*
*********************************************************************************************************
*                                                main()
*
* Description : This is the standard entry point for C code.  It is assumed that your code will call
*               main() once you have performed all necessary initialization.
*
* Argument(s) : none
*
* Return(s)   : none
*
* Caller(s)   : Startup Code.
*
* Note(s)     : none.
*********************************************************************************************************
*/
int  main (void)
{
    CPU_Init();                                                 /* Initialize the uC/CPU services                      */
    BSP_PreInit();                                              /* Preinit board support package                       */
    OSInit(&err);                                               /* Init uC/OS-III.                                     */
              
    OSTaskCreate((OS_TCB      *)&App_ControlTaskTCB,            /* Create the control task                             */
                 (CPU_CHAR    *)"Control",
                 (OS_TASK_PTR  )App_ControlTask, 
                 (void        *)0,
                 (OS_PRIO      )APP_CFG_TASK_CONTROL_PRIO,
                 (CPU_STK     *)&App_ControlTaskStk[0],
                 (CPU_STK_SIZE )APP_CFG_TASK_CONTROL_STK_SIZE_LIMIT,
                 (CPU_STK_SIZE )APP_CFG_TASK_CONTROL_STK_SIZE,
                 (OS_MSG_QTY   )0u,
                 (OS_TICK      )0u,
                 (void        *)0,
                 (OS_OPT       )(OS_OPT_TASK_STK_CHK | OS_OPT_TASK_STK_CLR),
                 (OS_ERR      *)&err);
    

    OSStart(&err);                                              /* Start multitasking (i.e. give control to uC/OS-III).*/
    
    (void)&err;
    
    return (0);
}


/*
*********************************************************************************************************
*                                          Control TASK
*
* Description : This task is responsible to control or output data to several devices like DC motor, stepper motor or SD card
*
* Arguments   : p_arg   is the argument passed to 'AppControlTask()' by 'OSTaskCreate()'.
*
* Returns     : Hopefully never :)
*
* Notes       : 1) The first line of code is used to prevent a compiler warning because 'p_arg' is not
*                  used.  The compiler should not generate any code for this statement.
*********************************************************************************************************
*/

static  void  App_ControlTask (void *p_arg)
{
    OS_ERR          err;    
    CPU_TS          ts;
    OS_MSG_SIZE     msg_size;
    (void)          p_arg;   
    void            *p_msg;

    unsigned char   writes=0;
      
    ControlMessage *message;
    unsigned int        bw;                                    /* bytes written */
    
    BSP_PostInit();                                             /* Initialize BSP functions  */   
   
    App_TaskCreate();                                           /* create additional tasks */
    App_ObjCreate();                                            /* create additional kernel objects */
    
    /* HW init */                   
    SPI_Init();                                                 /* SD card  */
   
    
    while (res != FR_OK)
    {
      FS_Init();                                                      /* Initialize FAT16/32 filesystem */
      res = f_open(&fsrc, "0:file.dat", FA_OPEN_ALWAYS | FA_WRITE);   /* Open file, create if it's not there */
      OSTimeDlyHMSM(0u, 0u, 2u, 0u,                                   /* wait a couple of seconds and try again if creation did not succeed */
                     OS_OPT_TIME_HMSM_STRICT, 
                      &err); 
    }
    
    res = f_lseek(&fsrc,fsrc.fsize);                                  /* position to end of file because we want to append data */
    res = f_write(&fsrc, "Starting controller\n", 20, &bw);           /* Write data to the file, Tell te log file that we've started */  
    
    while(DEF_TRUE)
    {           
                                                                      /* Wait for a control message to arrive */
          p_msg = OSQPend((OS_Q   *) &ControlQueue,
                  (OS_TICK) 0,
                  (OS_OPT) OS_OPT_PEND_BLOCKING,
                  (OS_MSG_SIZE *) &msg_size,
                  (CPU_TS *) &ts,
                  (OS_ERR   *) &err);          
          
                                                                      /* We've received a ControlMessage. Check contents */
         message = (ControlMessage *) p_msg;      
         res = f_write(&fsrc, timeStamp, strlen((const char *) timeStamp), &bw);   /* Write timestamp to the file */  
                
          switch (message->Cmd)
          {
              case  CMD_ROTATE:
                res = f_write(&fsrc, "Rotate\n", 7, &bw);                                 /* Write data to the file */  
                rotatePlatform();
                break;
              case  CMD_MOVE_UP:    
                res = f_write(&fsrc, "Move up\n", 8, &bw);                                /* Write data to the file */  
                stepUp();
                break;
              case  CMD_MOVE_DOWN:     
                res = f_write(&fsrc, "Move down\n", 10, &bw);                             /* Write data to the file */  
                stepDown();
                break;                
              case  CMD_STORE_LOG_MESSAGE:   
              case  CMD_STORE_GPS_LOCATION:       
                if ((strlen((const char *) message->Data) > 2))
                {
                  res = f_write(&fsrc, message->Data, strlen((const char *) message->Data), &bw);    /* Write log data to the file */  
                  writes++;
                 
                  if (res != FR_OK)                                                       /* if something went wrong.. what can we do?.. toggle led in this case */
                     BSP_LED_Toggle(5);
                }
                else
                      BSP_LED_Toggle(4);
                break;
              default:
                break;
          }
          message->MessageHandled = true;                                                 /* Set handled flag to true so message can be re-used again */
          BSP_LED_Toggle(0);    
          
          if (writes > 10)                                                                /* flush after 10 write commands */
          {
                res = f_sync(&fsrc);
                writes = 0;
          }

    }
}


/*
*********************************************************************************************************
*                                          AppTaskCreate()
*
* Description : Create application tasks.
*
* Argument(s) : none
*
* Return(s)   : none
*
* Note(s)     : none.
*********************************************************************************************************
*/

static  void  App_TaskCreate (void)
{
 
    OSTaskCreate((OS_TCB      *)&App_MonitorTaskTCB,              /* Create the Monitor task  */
                 (CPU_CHAR    *)"Monitor",
                 (OS_TASK_PTR  )App_MonitorTask, 
                 (void        *)0,
                 (OS_PRIO      )APP_CFG_TASK_MONITOR_PRIO,
                 (CPU_STK     *)&App_MonitorTaskStk[0],
                 (CPU_STK_SIZE )APP_CFG_TASK_MONITOR_STK_SIZE_LIMIT,
                 (CPU_STK_SIZE )APP_CFG_TASK_MONITOR_STK_SIZE,
                 (OS_MSG_QTY   )0u,
                 (OS_TICK      )0u,
                 (void        *)0,
                 (OS_OPT       )(OS_OPT_TASK_STK_CHK | OS_OPT_TASK_STK_CLR),
                 (OS_ERR      *)&err);
}


/*
*********************************************************************************************************
*                                          App_ObjCreate()
*
* Description : Create application kernel objects tasks.
*
* Argument(s) : none
*
* Return(s)   : none
*
* Note(s)     : none.
*********************************************************************************************************
*/

static  void  App_ObjCreate (void)
{

    OSQCreate((OS_Q *) &ControlQueue,
              (CPU_CHAR *)"Controll queue",
              (OS_MSG_QTY ) 10,
              (OS_ERR    *) &err);
}



/*
*********************************************************************************************************
*                                          FS_Init()
*
* Description : Initializes the FAT16/32 filesystem. Currently it's only mounting the device.
*
* Argument(s) : none
*
* Return(s)   : none
*
* Note(s)     : none.
*********************************************************************************************************
*/
static void FS_Init()
{
      /* Mount drive 0 */  
      res = f_mount(0, &fs);    
}

/*
*********************************************************************************************************
*                                          get_fattime()
*
* Description : Support function for the FAT16/32 filesystem. It returns the current date/time used at file operations
*
* Argument(s) : none
*
* Return(s)   : Time coded in format that the FAT16/32 understands :)
*
* Note(s)     : none.
*********************************************************************************************************
*/
DWORD get_fattime (void)
{
  
  return	          ((DWORD)(dateTime.Year - 1980) << 25)	                /* Y */
			| ((DWORD)dateTime.Month  << 21)			/* M */
			| ((DWORD)dateTime.Day << 16)				/* D */
			| ((DWORD)dateTime.Hour  << 11)				/* H */
			| ((DWORD)dateTime.Min  << 5)				/* M */
			| ((DWORD)dateTime.Sec  >> 1);				/* S */
}


/*
*********************************************************************************************************
*                                          rotatePlatform()
*
* Description : Rotates the platform about 3.5 degrees
*
* Argument(s) : none
*
* Return(s)   : none
*
* Note(s)     : none.
*********************************************************************************************************
*/
static void rotatePlatform()
{        
        /* start square wave on TO05 out */  
        if (!pwmEnabled)
        {
          pwmEnabled = true;
          startPWM();             
        }       
}

/*
*********************************************************************************************************
*                                          stepUp()
*
* Description : moves the platform up with approx. 1 degree
*
* Argument(s) : none
*
* Return(s)   : none
*
* Note(s)     : none.
*********************************************************************************************************
*/
static void stepUp()
{
     unsigned char i;
     for (i=0;i<100;i++)
     {
       if (stepperPosition >= 3) stepperPosition = 0;       
       else stepperPosition++;       
   
       STEPPER_POSITION(steps[stepperPosition]); 

       /* hold postition for at least 20msec */
       delay_ms(30);
       
       
       /* disabled for know. Sometimes we don't get scheduled in again... ? */
       //OSTimeDlyHMSM(0u, 0u, 0u, STEPPER_HOLD_TIME * 2, 
       //              OS_OPT_TIME_HMSM_STRICT, 
       //              &err);   
       
     }
     STEPPER_POSITION(STEPPER_REST);                                            /* move to rest position to save power */
}

/*
*********************************************************************************************************
*                                          stepDown()
*
* Description : moves the platform down with approx. 1 degree
*
* Argument(s) : none
*
* Return(s)   : none
*
* Note(s)     : none.
*********************************************************************************************************
*/
static void stepDown()
{
     unsigned char i;
     for (i=0;i<100;i++)
     {
       if (stepperPosition == 0) stepperPosition = 3;       
       else stepperPosition--;       
  
       STEPPER_POSITION(steps[stepperPosition]); 

       /* hold position for at least 20msec */

      delay_ms(30);
       /* disabled for know. Sometimes we don't get scheduled in again... ? */
       //OSTimeDlyHMSM(0u, 0u, 0u, STEPPER_HOLD_TIME * 2, 
       //              OS_OPT_TIME_HMSM_STRICT, 
       //              &err);   
     }    
     STEPPER_POSITION(STEPPER_REST);                                           /* move to rest position to save power */
}